;;; k-dir.el -- Directory display-name handling for k-mode
;;
;; Copyright (C) 2018 Marc Munro
;;
;; This file is not part of GNU Emacs
;;
;; Compatibility: GNU Emacs 24.x onwards
;;
;; License: GPL V3
;;   https://www.gnu.org/licenses/gpl-3.0.html


(require 'k-core)

(defvar k-dir::dirs-alist nil
  "Alist of directory paths to names of the form:
    (PATH . NAME)

  Where:
    NAME is the display name for the directory.  This must be unique among
    all open directories, and will be based upon the basename of the
    directory path.  If a duplicate name would be generated by the
    addition of a new directory, a uniquify process will be run which
    will possibly modify NAME.")

(defvar k-dir::display-name-update-hook
  "Functions added to this hook will be called whenever a display name
  is modified.")

;; HELPER FUNCTIONS

(defun k-dir::basenames (alist)
  "Return the set of basenames of paths from (the cars of) ALIST."
  (delete-dups
   (mapcar
    (lambda (entry)
      (file-name-nondirectory (car entry)))
    alist)))

(defun k-dir::abbreviated-alist (alist)
  "Return a copy of ALIST where each path entry is replaced with an
  abbreviated version of the full path - eg with ~x replacing /home/x,
  etc."
  (mapcar
   (lambda (entry)
     (cons (abbreviate-file-name (car entry)) (cdr entry)))
   alist))

(defun k-dir::matching-basenames (alist basename)
  "Return the set of entries in ALIST whose paths match BASENAME (ie
  the basename of the path matches our basename argument)."
  (delq
   nil
   (mapcar
    (lambda (entry)
      (and (string= basename (file-name-nondirectory (car entry)))
	   entry))
    alist)))

(defun k-dir::setentry (alist abbrevdir value)
  "Set the cdr of the ALIST entry matching ABBREVDIR to VALUE, setting
  the value of UPDATED (a dynamically-scoped variable) to t if this is
  a significant update."
  (let ((match (assoc (expand-file-name abbrevdir) alist)))
    (when match
      (and (and (cdr match)
		(not (string= (cdr match) value)))
	   (setq updated t))
      (setcdr match value))))

(defun k-dir::common-items (list1 list2)
  "Return a list of the initial elements in LIST1 that match LIST2."
  (if (string= (car list1) (car list2))
      (cons (car list1)
	    (and (cdr list1)
		 (k-dir::common-items (cdr list1) (cdr list2))))))

(defun k-dir::common-bases (paths)
  "Return a list containing the common elements of PATHS, a list of lists
  of directory names.  Eg if list is '((a b c) (a b d)), the result
  will be '(a b)." 
  (if (cdr paths)
      (k-dir::common-items (car paths)
			   (k-dir::common-bases (cdr paths)))
    (car paths)))

(defun k-dir::shared-suffix-elements (path paths)
  "Return the number of suffix elements from path that are shared with
  other path entries."
  (apply 'max
	 (mapcar
	  (lambda (this-path)
	    (if (eq path this-path)
		0
	      (length (k-dir::common-bases
		       (list (reverse path) (reverse this-path))))))
	  paths)))

(defun k-dir::attempt-prefix-removal (paths)
  "For each path in PATHS (a list of list of directory parts), attempt
  to remove as much of the prefix as possible, and return the modified
  list.  The definition of as much as possible is:
  - at least one entry must be left
  - whatever is left must not match the suffix of any other entry"
  (mapcar
   (lambda (path)
     (let ((target-len
	    (1+ (k-dir::shared-suffix-elements path paths))))
       (nthcdr (- (length path) target-len) path)))
   paths))

(defun k-dir::truncate-homedir-entry (paths alist)
  "For each path in PATHS that is a reference to our home directory,
  remove that home directory entry."
  (let* ((homedir (getenv "HOME"))
	 (len (length homedir))
	 (parts (cdr (split-string homedir "/"))))
    (cl-mapcar
     (lambda (entry1 entry2)
       (when (eq t (compare-strings homedir 0 len (car entry2) 0 len))
	 ;; This entry is for the user's home directory.  Remove any
	 ;; parts of homedir that have not already been removed.
	 ;; Note that this is a simplistic approach and may fail to
	 ;; deal with pathalogical cases.
	 
	 (dolist (part parts)
	   (if (string= part (car entry1))
	       (setq entry1 (cdr entry1)))))
       (or entry1 '("")))
     paths alist)))

(defun k-dir::shorten-intermediate-suffices (paths)
  "For each path in PATHS (a list of list of directory parts), replace
  the cdr of the list with a '...' entry."
  (mapcar
   (lambda (path)
     (if (cdr path)
	 (list (car path) "...")
       path))
   paths))

(defun k-dir::replace-suffices (paths suffices)
  "For each path in PATHS (a list of list of directory parts), replace
  append a contracted version of SUFFICES (which is actually in reverse
  order)."
  (let ((suffix
	 (if (> (length suffices) 1)
	     (list "..." (car suffices))
	   (list (car suffices)))))
    (mapcar
     (lambda (path)
       (append path suffix))
     paths)))

(defun k-dir::join-path-parts (paths)
  "Concatenate the components of each path in PATHS and eliminate any
  superfluous elipses, ie those at the start of the directory path, or
  that are duplicated."
  (mapcar
   (lambda (path)
     (replace-regexp-in-string
      "\\(\\(.../\\)\\|^/\\).../" "\\2"
      (mapconcat (lambda (x) x) path "/"))
     )
   paths))

(defun k-dir::distinguish-from-root (paths alist)
  "For each path in PATHS, attempt to update any entry that begins with
  '/', adding back the smallest possible prefix from ALIST in order to
  distinguish it from a real root entry." 
  (mapcar
   (lambda (path)
     (if (string= (substring path 0 1) "/")
	 ;; Path looks like root.
	 (let* ((path-len (length (split-string path "/")))
		(alist-parts (split-string (caar alist) "/"))
		(alist-len (length alist-parts)))
	   (setq path (concat (nth (- alist-len path-len) alist-parts) path))))
     (setq alist (cdr alist))
     path)
   paths))

(defun k-dir::uniquify-matching-basenames (alist basename)
  "Uniquify the display names in ALIST for paths whose basename is
  BASENAME, setting the dynamically-scoped UPDATED value to t if this
  results in display-names being changed.  See
  `k-dir::uniquify-display-names' for a description of uniquification."
  (let* ((abalist (k-dir::abbreviated-alist alist))
	 (matches (k-dir::matching-basenames abalist basename)))
    (if (= 1 (length matches))
	;; There is a single match so make the project name the
	;; basename.
	(k-dir::setentry alist (caar matches) basename)
      (let* ((paths
	      (mapcar
	       (lambda (match)
		 (reverse (split-string (car match) "/")))
	       matches))
	     (suffices (k-dir::common-bases paths))
	     (length (length suffices)))
	;;(message "PATHS: %s" paths)
	;;(message "SUFFICES: %s" suffices)
	;; 1) Remove common suffices from paths
	(setq paths (mapcar (lambda (rpath) (nthcdr length rpath)) paths))
	;;(message "PATHS(1): %s" paths)
	;; 2) Remove common prefixes from paths
	(setq paths (mapcar 'reverse paths))
	(setq length (length (k-dir::common-bases paths)))
	(setq paths (mapcar (lambda (path) (nthcdr length path)) paths))
	;;(message "PATHS(2): %s" paths)
	;; 3) Remove as much of other prefixes as possible
	(setq paths (k-dir::attempt-prefix-removal paths))
	;;(message "PATHS(3): %s" paths)
	;; 4) Remove homedirs
	(setq paths (k-dir::truncate-homedir-entry paths abalist))
	;;(message "PATHS(4): %s" paths)
	;; 5) Replace remaining suffices with "..."
	(setq paths (k-dir::shorten-intermediate-suffices paths))
	;;(message "PATHS(5): %s" paths)
	;; 6) Replace original suffices
	(setq paths (k-dir::replace-suffices paths suffices))
	;;(message "PATHS(6): %s" paths)
	;; 7) Concatenate parts and eliminate superfluous elipses
	(setq paths (k-dir::join-path-parts paths))
	;;(message "PATHS(7): %s" paths)
	;; 8) If any entry begins with "/", try adding back the smallest
	;; part of the prefix
	(setq paths (k-dir::distinguish-from-root paths abalist))
	;;(message "PATHS(8): %s" paths)
	;; Update the name part of each alist
	(cl-mapcar
	 (lambda (match displayname)
	   (k-dir::setentry alist (car match) displayname))
	 matches paths)
	))))

(defun k-dir::uniquify-display-names (alist &optional update-hook)
  "Ensure all display names in ALIST are unique, running the hook
  functions in UPDATE-HOOK if any display-names are modified.

  The rules for uniquification are as follows:
  1) Remove any common suffices (and retain them)
  2) Remove any common prefices
  3) Remove as much of each remaining prefix as possible while ensuring
     uniqueness
  4) Truncate references to our home directory
  5) Replace suffices of remaining values with /... to indicate removed
     directories 
  6) Add back abbreviated (with /...) versions of suffices removed in
     step 1
  7) Eliminate superflous /... sequences.

  Eg using: /root/proj/xyz/skit /home/karen/proj/xyz/skit 
            /home/marc/proj/xyz/skit /home/marc/xyz/skit
  1) ==> /root/proj /home/karen/proj /home/marc/proj /home/marc
  2) ==> root/proj home/karen/proj home/marc/proj home/marc
  3) ==> root/proj karen/proj marc/proj marc
  4) ==> root/proj karen/proj proj <empty string>
  4) ==> root/... karen/... proj <empty string>
  6) ==> root/.../.../skit karen/.../.../skit proj/.../skit .../skit
  7) ==> root/.../skit karen/.../skit proj/.../skit skit"

  (let ((basenames (k-dir::basenames alist))
	updated)
    (mapc
     (lambda (basename)
       (k-dir::uniquify-matching-basenames alist basename))
     basenames)
    (if (and updated update-hook)
	(run-hooks update-hook))))

(defun k-dir::record-path-in-alist (path alist-sym update-hook)
  "Ensure PATH is present in the variable given by ALIST-SYM,
  adding it and re-uniquifying the alist if needed.  Functions in
  UPDATE-HOOK will be run by the uniquify process if any display names
  are modified."

  (unless (assoc path (symbol-value alist-sym))
    (set alist-sym (cons (cons path nil)
			 (symbol-value alist-sym)))
    (k-dir::uniquify-display-names (symbol-value alist-sym) update-hook)))
  
  
;; PUBLIC INTERFACE FUNCTIONS

(defun k-dir::record-dir (path)
  "Ensure that PATH is known, recording its existence and
  re-uniquifying all dir names if it is not."
  (k-dir::record-path-in-alist
   path 'k-dir::dirs-alist 'k-dir::display-name-update-hook))

(defun k-dir::dirname (path)
  "Return the display name for the directory given by PATH."
  (cdr (or (assoc path k-dir::dirs-alist)
	   (progn
	     (k-dir::record-dir path)
	     (assoc path k-dir::dirs-alist)))))

(defun k-dir::is-only-buffer-for-dir (buf)
  "T if BUF is the only buffer for its directory."
  (let ((buffer-dir (k::buffer-dir buf)))
    (and buffer-dir
	 (catch 'found-match
	   (mapc
	    (lambda (this-buf)
	      (unless (eq buf this-buf)
		(and (string= buffer-dir (k::buffer-dir this-buf))
		     (throw 'found-match nil)
		     )))
	    (buffer-list))
	   t))))

(defun k-dir::maybe-forget-dir-for-buffer (buf)
  "Check all current buffers to determine whether buffer BUF is the
  only buffer for its directory.  This should be called prior to
  deleting the buffer."
  (if (k-dir::is-only-buffer-for-dir buf)
      (let* ((dir (k::buffer-dir buf))
	     (match (assoc dir k-dir::dirs-alist)))
	(setq k-dir::dirs-alist (delq match k-dir::dirs-alist))
	(k-dir::uniquify-display-names
	 k-dir::dirs-alist
	 'k-dir::display-name-update-hook))))
	
(provide 'k-dir)

;;(k-dir::dirname "/home/marc/emacs/k")
;;(k-dir::dirname "/home/marc/k")
;;k-dir::dirs-alist
;;(setq k-dir::dirs-alist nil)
;;(k-dir::maybe-forget-dir-for-buffer (get-buffer "x"))
