;;; k-project.el --- Project identification and handling for k-mode
;;
;; Copyright (C) 2019 Marc Munro
;;
;; This file is not part of GNU Emacs
;;
;; Compatibility: GNU Emacs 24.x onwards
;;
;; License: GPL V3
;;   https://www.gnu.org/licenses/gpl-3.0.html

;; K-project
;; Stuff for associating buffers with projects
;;
;; A buffer belongs to a project iff it represents a file or directory
;; within a project directory.   The project directory for a given
;; directory path, is returned by `k::vcs-dir-for-path'.
;;

(require 'k-core)
(require 'k-dir)

(defvar k-project::projects-alist nil
  "Alist of project paths to names of the form:
    (PATH . NAME)

  Where:
    NAME is the display name for the project.  This must be unique among
    all projects, and will be based upon the basename of the project
    path.  If a duplicate name would be generated by the addition of a new
    project, a uniquify process will be run which will possibly modify
    NAME.")

(defvar k-project::display-name-update-hook
  "Functions added to this hook will be called whenever a display name
  is modified.")


(defun k-project::project-name (path)
  "Search up the directory tree looking for a vcs folder."
  (let ((ppath (k::vcs-dir-for-path path))
	(pentry))
    (when ppath
      (setq pentry (or (assoc ppath k-project::projects-alist)
		       (k-project::record-project ppath)))
      (cdr pentry))))

(defun k-project::record-project (path)
  "Add a new project PATH to `k-project::projects-alist', ensuring that
its name is unique."
  (k-dir::record-path-in-alist
   path 'k-project::projects-alist 'k-project::project-name-update-hook))

(defun k-project::buffer-project (&optional buffer)
  "Return the project for BUFFER (or `current-buffer')."
  (k-project::project-name (k::project-path buffer)))
  
(defun k-project::project-buffer-p ()
  "Identify whether the current buffer belongs to the project defined
in dynamically-scoped variable `k::project'."
  (string= k::project (k-project::buffer-project))) 

(defvar k-project::save-project-buffers-only nil
  "Whether `k-project::save-some-buffers' should attempt to save only
  project buffers.  This should be overridden in local let bindings.")

(defun k-project::save-some-buffers (orig-fun &optional arg predicate)
  "Save modified buffers only for files in the project of
`current-buffer'."
  (let ((k::project (k-project::buffer-project)))
    (if k::project
	(setq predicate
	      (lambda ()
		(string= k::project (k-project::buffer-project)))))
    (apply orig-fun arg (list predicate))))

(defun k-project::is-only-buffer-for-project (buf)
  "T if buf is the only buffer for its project."
  ;; TODO: Consider refactoring this with the equivalent function in k::dir
  (let ((project (k-project::buffer-project buf)))
    (and project
	 (catch 'found-match
	   (mapc
	    (lambda (this-buf)
	      (unless (eq buf this-buf)
		(and (string= project (k-project::buffer-project this-buf))
		     (throw 'found-match nil))))
	    (buffer-list))
	   t))))

(defun k-project::maybe-forget-project-for-buffer (buf)
  "Check all current buffers to determine whether buffer BUF is the
  only buffer for its project.  This should be called prior to
  deleting the buffer."
  (if (k-project::is-only-buffer-for-project buf)
      (let* ((project (k-project::buffer-project buf))
	     (match (assoc project k-project::projects-alist)))
	(setq k-project::projects-alist (delq match k-project::projects-alist))
	(k-dir::uniquify-display-names
	 k-project::projects-alist
	 'k-project::display-name-update-hook))))


(defun k-project::kill-buffer ()
  "Hook function called before the current buffer is killed.
  Responsible for dealing with the removal of projects
  from `k-project::projects-alist'."
  (let ((buffer (current-buffer)))
    (k-project::maybe-forget-project-for-buffer buffer)))


;;;
;;; ACTIVATION
;;;

(defun k-project::activate ()
  "Activate k-project stuff."
  (add-hook 'kill-buffer-hook #'k-project::kill-buffer)
  (advice-add 'save-some-buffers :around #'k-project::save-some-buffers))
  
(defun k-project::deactivate ()
  "Deactivate k-project stuff."
  (remove-hook 'kill-buffer-hook #'k-project::kill-buffer)
  (advice-remove 'save-some-buffers #'k-project::save-some-buffers))
  
(add-hook 'k-mode-activation-hook #'k-project::activate)
(add-hook 'k-mode-deactivation-hook #'k-project::deactivate)

(provide 'k-project)

