;; k-tabbar.el --- tabbar stuff for k-mode 
;;
;; Copyright (C) 2019 Marc Munro
;;
;; This file is not part of GNU Emacs
;;
;; Compatibility: GNU Emacs 26.x
;;
;; License: GPL V3
;;   https://www.gnu.org/licenses/gpl-3.0.html
;;
;; buffer-group types:
;;  proj  minimally unique proj-path:
;;  dir   minimally unique directory-file-name
;;  mode  [mode name sans -mode suffix]
;;  proj+dir minimally unique proj-path: dir suffix
;;  proj+mode
;;

(require 'k-core)
(require 'k-dir)

;; CORE STUFF

(defun k-tabbar2::buffer-project (buffer)
  "Return the project path for the project for BUFFER, or nil."
  (and (fboundp #'k-project::project-name)
       (cdr (k2::project-path buffer))))

(defun k-tabbar2::buffer-mode (buffer)
  "Return the major mode of BUFFER, or nil if this is a special
  emacs buffer"
  (unless (k::emacs-buffer-p buffer)
    (with-current-buffer buffer major-mode)))

(defun k-tabbar2::sort-by-list(list ordered-list)
  "Return a list consisting of ((a) (b)) where (a) is a list of
   items from LIST in the order that they appear in ORDERED-LIST,
   and (b) is the set of items from LIST that do not appear in
   ordered-list."
  (let* ((list (copy-sequence list))
	 (found
	  (delq
	   nil
	   (mapcar
	    (lambda(x)
	      (when (memq x list)
		(setq list (delq x list))
		x))
	    ordered-list))))
    (list found list)))

(defun k-tabbar2::sort-from-list-to-list(list ordered-list)
  "Return a list of items from LIST in the order that they appear in
   ORDERED-LIST, followed by items from LIST that do no not appear in
   ORDERED-LIST."
  (let ((parts (k-tabbar2::sort-by-list list ordered-list)))
    (if (cadr parts)
	(append (car parts) (cadr parts)) ;; concat the lists
      (car parts))))

(defvar k-tabbar2::all-buffer-groups nil
  "List of all buffer-groups.  This is used to provide the definitive
  elements that identify a buffer group so that comparisons for
  buffer-groups can be performed with eq for performance reasons.")

(defun k-tabbar2::buffer-group (buffer-group)
  "Retrieve the definitive buffer-group element from
  k-tabbar2::all-buffer-groups, adding a new element if needed."
  (and buffer-group
       (car (or (member buffer-group k-tabbar2::all-buffer-groups)
		(setq k-tabbar2::all-buffer-groups
		      (cons buffer-group k-tabbar2::all-buffer-groups))))))

(defun k-tabbar2::delete-buffer-group (buffer-group)
  "Remove buffer-group from k-tabbar2::all-buffer-groups"
  (setq k-tabbar2::all-buffer-groups
	(delq buffer-group k-tabbar2::all-buffer-groups)))

(defun k-tabbar2::buffer-buffer-groups (buffer)
  "Return a list of all buffer-groups for BUFFER.
  This lists all possible buffer-groups that buffer may belong to,
  other than frame buffer groups.  This is a larger set than the set
  that may be active, which will depend on which windows have shown
  which buffer-groups."
  (and buffer
       (not (minibufferp buffer))
       (let ((proj (k-tabbar2::buffer-project buffer))
	     (mode (k-tabbar2::buffer-mode buffer))
	     (dir (k::buffer-dir buffer)))
	 (let ((p (k-tabbar2::buffer-group
		   (and proj (list 'proj proj))))
	       (p+d (k-tabbar2::buffer-group
		     (and proj dir (list 'proj+dir proj dir))))
	       (p+m (k-tabbar2::buffer-group
		     (and proj mode (list 'proj+mode proj mode))))
	       (p+d+m (k-tabbar2::buffer-group
		       (and proj dir mode (list 'proj+dir+mode proj dir
						mode))))
	       (d (k-tabbar2::buffer-group
		   (and dir (list 'dir dir))))
	       (d+m (k-tabbar2::buffer-group
		     (and dir mode (list 'dir+mode dir mode))))
	       (m (k-tabbar2::buffer-group
		   (and mode (list 'mode mode))))
	       (s (k-tabbar2::buffer-group
		   (and (not mode) (list 'emacs)))))
	   (delq nil
		 (list p p+d p+m p+d+m d d+m m s))))))

(defun k-tabbar2::frame-buffer-group (frame)
  "Return the frame buffer-group for FRAME."
  (k-tabbar2::buffer-group (list 'frame frame)))
  
(defun k-tabbar2::window-buffer-groups (window)
  "Return a list of possible buffer-groups for WINDOW, based
  on the window's currently selected buffer, including the window
  frame's own buffer group.  This is sorted so that more recently
  used buffer-groups appear before less recently used ones."
  (let ((buffer (window-buffer window))
	(shown-bgs (window-parameter window 'k-tabbar::buffer-groups)))
    (append (k-tabbar2::buffer-buffer-groups buffer)
	    (list (k-tabbar2::frame-buffer-group (window-frame window))))))

(defun k-tabbar2::proj-buffers (project-path)
  "Return all buffers that are part of the project given by
PROJECT-PATH."
  (and project-path
       (delq
	nil
	(mapcar
	 (lambda (buffer)
	   (and (equal project-path (k-tabbar2::buffer-project buffer))
		buffer))
	 (buffer-list)))))

(defun k-tabbar2::mode-buffers (mode &optional buffer-list)
  "Return all buffers that have major-mode MODE, from BUFFER-LIST."
  (and mode
       (delq nil
	     (mapcar
	      (lambda (buffer)
		(and (eq mode (k-tabbar2::buffer-mode buffer)) buffer))
	      (or buffer-list (buffer-list))))))

(defun k-tabbar2::dir-buffers (dir &optional buffer-list)
  "Return all buffers from directory DIR, from BUFFER-LIST."
  (and dir
       (delq nil
	     (mapcar
	      (lambda (buffer)
		(and (equal dir (k::buffer-dir buffer)) buffer))
	      (or buffer-list (buffer-list))))))

(defun k-tabbar2::remove-minibuffers-from-list (buffer-list)
  "Remove minibuffers from BUFFER-LIST."
  (delq nil
	(mapcar
	 (lambda (buffer)
	   (and (not (minibufferp buffer)) buffer))
	 buffer-list)))

(defun k-tabbar2::emacs-buffers (&optional buffer-list)
  "Return all emacs buffers, from BUFFER-LIST."
  (delq nil
	(mapcar
	 (lambda (buffer)
	   (and (k::emacs-buffer-p buffer) buffer))
	 (or buffer-list (buffer-list)))))

(defun k-tabbar2::buffer-group-buffers (buffer-group)
  "Return the list of buffers which are members of BUFFER_GROUP"
  (let ((buffer-list)
	(bg-type (car buffer-group)))
    (cond
     ((eq bg-type 'proj)
      (k-tabbar2::proj-buffers (cadr buffer-group)))
     ((eq bg-type 'mode)
      (k-tabbar2::mode-buffers (cadr buffer-group)))
     ((eq bg-type 'dir)
      (k-tabbar2::dir-buffers (cadr buffer-group)))
     ((eq bg-type 'proj+mode)
      (k-tabbar2::mode-buffers (nth 2 buffer-group)
			       (k-tabbar2::proj-buffers (cadr buffer-group))))
     ((eq bg-type 'proj+dir)
      (k-tabbar2::dir-buffers (nth 2 buffer-group)
			      (k-tabbar2::proj-buffers (cadr buffer-group))))
     ((eq bg-type 'proj+dir+mode)
      (k-tabbar2::dir-buffers
       (nth 2 buffer-group)
       (k-tabbar2::mode-buffers
	(nth 3 buffer-group)
	(k-tabbar2::proj-buffers (cadr buffer-group)))))
     ((eq bg-type 'emacs)
      (k-tabbar2::emacs-buffers))
     ((eq bg-type 'dir+mode)
      (k-tabbar2::mode-buffers (nth 2 buffer-group)
			       (k-tabbar2::dir-buffers
				(cadr buffer-group))))
     ((eq bg-type 'frame)
      (k-tabbar2::remove-minibuffers-from-list
       (frame-parameter (cadr buffer-group) 'buffer-list))))))
     

(defvar k-tabbar2::known-buffers nil
  "List of buffers known to k-tabbar2.  This is used when trying to
figure out if we have a new buffer.")

(defun k-tabbar2::init-known-buffers ()
  "Set k-tabbar2::known-buffers to contain all known buffers."
  (setq k-tabbar2::known-buffers (copy-tree (buffer-list))))

(defun k-tabbar2::init-cur-buffers ()
  "Set the cached cur-buffer value for each window."
  (walk-windows
   (lambda (window)
     (k-tabbar2::cache-set
      'k-tabbar2::cached-cur-buffer
      (window-buffer window)
      window))
   nil t))

(defun k-tabbar2::new-buffer-p(buffer)
  "Predicate returning whether BUFFER is new to us."
  (if (member buffer k-tabbar2::known-buffers)
      nil
    (setq k-tabbar2::known-buffers (cons buffer k-tabbar2::known-buffers))))
  
(defun k-tabbar2::calculate-frame-buffer-group-buffer-list
    (frame buffer-group)
  "Return a newly calculated sorted buffer-list for FRAME and
    BUFFER-GROUP.  This is sorted by how recently each buffer has been
    shown in the frame, or how deeply buried it is in emacs' buffer-list."
  (let ((buffer-group-buffers
	 (k-tabbar2::buffer-group-buffers buffer-group))
	partial-result
	remaining-result)
    ;; (car partial-result) will contain those buffer-group-buffers that
    ;; were found in the frame's buffer-list, sorted according to that
    ;; list.   (cadr partial-result) contains the remaining buffers,
    ;; which we will sort according to emacs buffer-list. 
    (setq partial-result
	  (k-tabbar2::sort-by-list
	   buffer-group-buffers
	   (frame-parameter frame 'buffer-list)))
    (setq remaining-result
	  (k-tabbar2::sort-by-list
	   (cadr partial-result) (buffer-list)))
    ;; Ensure all buffers are recorded as known buffers and return the
    ;; concatenation of partial-result and remaining-result
    (mapc
     (lambda (buf) (k-tabbar2::new-buffer-p buf))
     (append (car partial-result) (car remaining-result)))))


(defun k-tabbar2::record-buffer-group-for-window (window buffer-group)
  "Record that in WINDOW BUFFER-GROUP has been recently selected."
  (set-window-parameter
   window 'k-tabbar::buffer-groups
   (cons buffer-group
	 (delq buffer-group
	       (window-parameter window 'k-tabbar::buffer-groups)))))

(defun k-tabbar2::best-window-buffer-group (window buffer)
  "Return the most appropriate buffer-group for WINDOW."
  (let ((cur-bg (k-tabbar2::cache-get
		 'k-tabbar2::cached-buffer-group window))
	(buffer-bgs (k-tabbar2::buffer-buffer-groups buffer))
	result in-buffer-bgs)
    (if (eq (car cur-bg) 'frame)
	;; The current buffer group is a frame buffer-group, so if
	;; buffer has been displayed on this frame, then it is the
	;; best window buffer-group
	(if (memq buffer
		  (frame-parameter (window-frame window) 'buffer-list))
	    (setq result cur-bg)))
    (cond
     (result
      result)
     ((memq cur-bg buffer-bgs)
      cur-bg)
     ((setq in-buffer-bgs (memq cur-bg buffer-bgs))
      (car in-window-bgs))
     (t
      (car buffer-bgs)))))
  
(defun k-tabbar2::buffer-group-known-to-frame (buffer-group frame)
  "Predicate to determine whether BUFFER-GROUP contains buffers that
have been displayed on FRAME."
  (catch 'found
    (let ((frame-buffers (frame-parameter frame 'buffer-list)))
      (mapc
       (lambda (buffer)
	 (if (memq buffer frame-buffers)
	     (throw 'found t)))
       (k-tabbar2::buffer-group-buffers buffer-group))
      nil)))
  
(defvar k-tabbar2::buffer-entry-keymap
  (let ((keymap (make-sparse-keymap)))
    (define-key keymap [header-line down-mouse-1]
      'k-tabbar2::select-buffer)
    (define-key keymap [header-line drag-mouse-1]
      'k-tabbar2::buffer-drag)
    keymap)
  "The keymap for buffer selection buttons."
  )

(defvar k-tabbar2::buffer-group-entry-keymap
  (let ((keymap (make-sparse-keymap)))
    (define-key keymap [header-line down-mouse-1]
      'k-tabbar2::select-buffer-group)
    (define-key keymap [header-line C-mouse-1]
      'k-tabbar2::act-on-buffer-entry)
    (define-key keymap [header-line M-mouse-1]
      'k-tabbar2::act-on-buffer-entry)
    (define-key keymap [header-line down-mouse-2]
      'k-tabbar2::act-on-buffer-entry)
    (define-key keymap [header-line C-mouse-2]
      'k-tabbar2::act-on-buffer-entry)
    (define-key keymap [header-line M-mouse-2]
      'k-tabbar2::act-on-buffer-entry)
    (define-key keymap [header-line down-mouse-3]
      'k-tabbar2::act-on-buffer-entry)
    (define-key keymap [header-line C-mouse-3]
      'k-tabbar2::act-on-buffer-entry)
    (define-key keymap [header-line M-mouse-3]
      'k-tabbar2::act-on-buffer-entry)
    keymap)
  "The keymap for buffer-group selection buttons."
  )

(defvar k-tabbar2::buffer-group-keymap
  (let ((keymap (make-sparse-keymap)))
    (define-key keymap [header-line mouse-1]
      'k-tabbar2::toggle-tabbar-display-type)
    (define-key keymap [header-line C-mouse-1]
      'k-tabbar2::toggle-tabbar-display-type)
    (define-key keymap [header-line M-mouse-1]
      'k-tabbar2::toggle-tabbar-display-type)
    (define-key keymap [header-line mouse-2]
      'k-tabbar2::toggle-tabbar-display-type)
    (define-key keymap [header-line C-mouse-2]
      'k-tabbar2::toggle-tabbar-display-type)
    (define-key keymap [header-line M-mouse-2]
      'k-tabbar2::toggle-tabbar-display-type)
    (define-key keymap [header-line mouse-3]
      'k-tabbar2::toggle-tabbar-display-type)
    (define-key keymap [header-line C-mouse-3]
      'k-tabbar2::toggle-tabbar-display-type)
    (define-key keymap [header-line M-mouse-3]
      'k-tabbar2::toggle-tabbar-display-type)
    (define-key keymap [header-line mouse-4] 'ignore)
    (define-key keymap [header-line mouse-4] 'ignore)
    (define-key keymap [header-line mouse-5] 'ignore)
    keymap)
  "The keymap for the leftmost buffer-group button."
  )

(defvar k-tabbar2::scroll-button-keymap
  (let ((keymap (make-sparse-keymap)))
    keymap)
  "The keymap for scroll buttons"
  ;; TODO: Implement this properly
  )

(defconst k-tabbar2::selected-text-color "blue"
  "Color of text for the selected buffer and buffer-group in a window.")

(defconst k-tabbar2::mouse-selected-text-color "blue"
  "Color of mouseover text for the selected buffer and buffer-group in a
  window.") 

(defconst k-tabbar2::mouse-default-text-color "DarkViolet"
  "Color of mouseover text for an unselected element in a window.")

(defconst k-tabbar2::scroll-button-text-color "black"
  "Color of text for scroll buttons.")

(defconst k-tabbar2::scroll-button-selected-text-color "blue"
  "Color of text for scroll buttons.")

(defconst k-tabbar2::default-text-color "grey20"
  "Color of text for non-selected items.")

(defconst k-tabbar2::undisplayed-text-color "grey50"
  "Color of text for buffers that have never been displayed in the
  current window.")

(defconst k-tabbar2::buffer-background-color "grey80"
  "Color of background for buffer buttons.")

(defconst k-tabbar2::buffer-group-background-color "PaleGoldenrod"
  "Color of background for buffer-group buttons.")

(defconst k-tabbar2::button-size 0.8
  "Size of (text inside of) buttons.")

(defun k-tabbar2::element-text-color (is-selected known-to-frame)
  "Return the color to display the text in the tabbar for an element
  based on IS-SELECTED and KNOWN-TO_FRAME."
  (if is-selected
      k-tabbar2::selected-text-color
    (if known-to-frame
	k-tabbar2::default-text-color
      k-tabbar2::undisplayed-text-color)))

(defun k-tabbar2::mouse-text-color (is-selected)
  "Return the color to display element text in the tabbar when the entry
  is moused-over."
  (if is-selected
      k-tabbar2::mouse-selected-text-color
    k-tabbar2::mouse-default-text-color))

(defun k-tabbar2::buffer-description (buffer)
  "Return a description of BUFFER suitable for use in the help text in a
tabbar."
  (let ((project-info (k2::project-info buffer))
	(mode (k-tabbar2::buffer-mode buffer))
	(path (k::buffer-dir buffer))
	(name (buffer-name buffer)))
    (concat
     (format "    Buffer: %s\n\n" name)
     (if project-info
	 (format "   Project: %s (%s)\n"
		 (cdr (assq 'path project-info))
		 (cdr (assq 'vcs project-info)))
       "")
     (if path (format " Directory: %s\n" path) "")
     (format "Major Mode: %s" mode))))

(defun k-tabbar2::buffer-group-project (buffer-group)
  "Return the path of BUFFER-GROUP's project."
  (let ((bgtype (car buffer-group)))
    (and (or (eq bgtype 'proj)
	     (eq bgtype 'proj+dir)
	     (eq bgtype 'proj+mode)
	     (eq bgtype 'proj+dir+mode))
	 (cadr buffer-group))))

(defun k-tabbar2::buffer-group-dir (buffer-group)
  "Return the path of BUFFER-GROUP's dir."
  (let ((bgtype (car buffer-group)))
    (cond
     ((or
       (eq bgtype 'dir)
       (eq bgtype 'dir+mode))
      (cadr buffer-group))
     ((or (eq bgtype 'proj+dir)
	  (eq bgtype 'proj+dir+mode))
      (nth 2 buffer-group)))))

(defun k-tabbar2::buffer-group-mode (buffer-group)
  "Return BUFFER-GROUP's mode."
  (let ((bgtype (car buffer-group)))
    (cond
     ((eq bgtype 'mode)	(cadr buffer-group))
     ((or (eq bgtype 'proj+mode)
	  (eq bgtype 'dir+mode))
      (nth 2 buffer-group))
     ((eq bgtype 'proj+dir+mode) nil
      (nth 3 buffer-group)))))

(defun k-tabbar2::buffer-group-description (buffer-group)
  "Return a description of BUFFER-GROUP suitable for use in the help
  text in a tabbar."
  (let* ((bgtype (car buffer-group))
	 (project (k-tabbar2::buffer-group-project buffer-group))
	 (vcs-info (and project
			(k2::project-info-from-path-info
			 (k2::vcs-info-for-path project))))
	 (proj-description (and project
				(format "%s (%s)"
					(cdr (assq 'path vcs-info))
					(cdr (assq 'vcs vcs-info)))))
	 (dir (k-tabbar2::buffer-group-dir buffer-group))
	 (mode (k-tabbar2::buffer-group-mode buffer-group)))
    (cond
     ((eq bgtype 'proj)
      (format
       (concat"        Project Buffer Group\n\n"
	      "    For project: %s")
       proj-description))
     ((eq bgtype 'proj+dir)
      (format
       (concat"   Project and Directory Buffer Group\n\n"
	      "    For project: %s,\n"
	      "      directory: %s")
       proj-description dir))
     ((eq bgtype 'proj+dir+mode)
      (format
       (concat"Project, Directory and Mode Buffer Group\n\n"
	      "    For project: %s,\n"
	      "      directory: %s,\n"
	      "     major mode: %s")
       proj-description dir mode))
     ((eq bgtype 'proj+mode)
      (format
       (concat"     Project and Mode Buffer Group\n\n"
	      "    For project: %s,\n"
	      "     major mode: %s")
       proj-description mode))
     ((eq bgtype 'dir)
      (format
       (concat"      Directory Buffer Group\n\n"
	      "  For directory: %s")
       dir))
     ((eq bgtype 'dir+mode)
      (format
       (concat"   Directory and Mode Buffer Group\n\n"
	      "  For directory: %s,\n"
	      "     major mode: %s")
       dir mode))
     ((eq bgtype 'mode)
      (format
       (concat"       Mode Buffer Group\n\n"
	      " For major mode: %s")
       mode))
     ((eq bgtype 'emacs)
      (concat" Buffer Group for emacs special buffers\n\n"
	     "              *emacs*"))
     ((eq bgtype 'frame)
      (format
       (concat "      Frame Buffer Group\n\n"
	       " For frame: %s")
       (cdr buffer-group))))))

(defvar k-tabbar2::project-name-face
  (list
   :height 1.1
   :weight 'bold)
  "Face for the name part of a project name and directory combination.")

(defun k-tabbar2::project-name (path)
  "Return a propertized representation of the unique project name for
  PATH."
  (let* ((name (k-project::project-name path))
	 (prefix (replace-regexp-in-string "[^/]*$" "" name))
	 (suffix (replace-regexp-in-string ".*/" "" name)))
    (concat prefix suffix)))

(defun k-tabbar2::mode-name (posn buffer-group)
  "Return the mode-name from element POSN within BUFFER-GROUP as a
  string."
  (replace-regexp-in-string
   "-mode$" ""
   (symbol-name (nth posn buffer-group))))

(defun k-tabbar2::buffer-group-name (buffer-group)
  "Return the name of BUFFER-GROUP suitable for use in an element button."
  (let* ((bgtype (car buffer-group))
	 (project (k-tabbar2::buffer-group-project buffer-group))
	 (vcs-info (and project
			(k2::project-info-from-path-info
			 (k2::vcs-info-for-path project))))
	 (project-name (and project
			    (k-tabbar2::project-name project)))
	 (project-prefix
	  (format "%s:%s" (cdr (assq 'vcs vcs-info)) project-name)))
    (cond
     ((eq bgtype 'proj)
      project-prefix)
     ((eq bgtype 'proj+dir)
      (format "%s%s/" project-prefix
	      (substring (nth 2 buffer-group) (length project))))
     ((eq bgtype 'proj+mode)
      (format "%s(%s)" project-prefix (k-tabbar2::mode-name 2 buffer-group)))
     ((eq bgtype 'proj+dir+mode)
      (format "%s%s/(%s)" project-prefix
	      (substring (nth 2 buffer-group) (length project))
	      (k-tabbar2::mode-name 3 buffer-group)))
     ((eq bgtype 'dir)
      (concat (k-dir::dirname (cadr buffer-group)) "/"))
     ((eq bgtype 'dir+mode)
      (format "%s/(%s)" 
	      (k-dir::dirname (cadr buffer-group))
	      (k-tabbar2::mode-name 2 buffer-group)))
     ((eq bgtype 'mode)
      (format "(%s)" 
	      (k-tabbar2::mode-name 1 buffer-group)))
     ((eq bgtype 'emacs)
      "*emacs*")
     ((eq bgtype 'frame)
      "[Frame]")
     )))



(defun k-tabbar2::face (height fg-color bg-color pressed &optional weight)
  "Defines a face for displaying things in the tabbar."
  (if height
      (if weight
	  (list
	   :inherit 'variable-pitch
	   :height height
	   :background bg-color
	   :foreground fg-color
	   :weight weight
	   :box `(:color "white"
			 :style ,(if pressed 'pressed-button 'released-button)))

	(list
	 :inherit 'variable-pitch
	 :height height
	 :background bg-color
	 :foreground fg-color
	 :box `(:color "white"
		       :style ,(if pressed 'pressed-button 'released-button))))
    (list
     :inherit 'variable-pitch
     :background bg-color
     :foreground fg-color
     :box `(:color "white"
		   :style ,(if pressed 'pressed-button 'released-button)))))

(defun k-tabbar2::separator(bgcolor)
  "Return a propertized string implementing a separator for insertion
between tabbar buttons."
  (propertize
   " "
   'element-type 'separator
   'font-lock-face (list :inherit 'variable-pitch
			 :height 1
			 :background bgcolor)
   'face (list :inherit 'variable-pitch
	       :height 1
	       :background bgcolor)))

(defun k-tabbar2::element-button
    (label help keymap fg-face mouse-face &rest properties)
  "Return a propertized string implementing a button for a buffer or
  buffer-group (an element as opposed to a separator or scroll button).
  LABEL is the text that appears in the button;
  HELP is the mouseover (help) text for the button;
  KEYMAP is the button's keymap, providing its functionality;
  FG-FACE is the face defintion for the button;
  MOUSE-FACE is the face definition for the pointer hovering on the button;
  PROPERTIES are any other properties that should be added."
  (apply
   #'propertize
   label
   'pointer 'hand
   'help-echo help
   'local-map keymap
   'font-lock-face fg-face
   'face fg-face
   'mouse-face mouse-face
   'element-type 'element
   properties))

(defun k-tabbar2::scroll-button (direction bg-color)
  "Return a scroll-button for DIRECTION (which will should be '< or
  '>) with a background color of BG-COLOR."
  ;; TODO: remove let bindings (replace by inline code) after testing
  (let ((face (k-tabbar2::face
	       k-tabbar2::button-size
	       k-tabbar2::scroll-button-text-color
	       bg-color nil 'ultra-bold))
	(mouse-face (k-tabbar2::face
		     nil
		     k-tabbar2::scroll-button-selected-text-color
		     bg-color nil 'ultra-bold))
	(text (if (eq direction '<) " < " " > "))) ;;"«" "»"

    (apply
     #'propertize
     text
     'pointer 'hand
     ;; TODO: add some scroll-button help here
     ;;'help-echo help
     'local-map k-tabbar2::scroll-button-keymap
     'font-lock-face face
     'face face
     'mouse-face mouse-face
     'element-type 'scroll-button
     nil)))

(defun k-tabbar2::buffer-button (buffer window)
  "Return a tabbar-displayable string representation of BUFFER in WINDOW."
  (let* ((is-selected (eq (window-buffer window) buffer))
	 (frame (window-frame window))
	 (known-to-frame (and (memq
			       buffer
			       (frame-parameter frame 'buffer-list))
			      t))
	(description (k-tabbar2::buffer-description buffer)))
    (k-tabbar2::element-button
     (concat " " (buffer-name buffer) " ")
     description
     k-tabbar2::buffer-entry-keymap
     (k-tabbar2::face
      k-tabbar2::button-size
      (k-tabbar2::element-text-color is-selected known-to-frame)
      k-tabbar2::buffer-background-color nil)
     (k-tabbar2::face
      nil
      (k-tabbar2::mouse-text-color is-selected)
      k-tabbar2::buffer-background-color t)
     'buffer buffer
     )))

(defun k-tabbar2::buffer-group-button (keymap buffer-group window)
  "Return a tabbar-displayable string button with functions from KEYMAP,
  representing BUFFER-GROUP in WINDOW."
  (let* ((is-selected (eq (k-tabbar2::get
			   'k-tabbar2::cached-buffer-group window)
			  buffer-group))
	 (frame (window-frame window))
	 (known-to-frame
	  (or (k-tabbar2::buffer-group-known-to-frame
	       buffer-group frame)))
	 (text-color (k-tabbar2::element-text-color
		      is-selected known-to-frame))
	 (mouse-text-color (k-tabbar2::mouse-text-color is-selected))
	 (description (k-tabbar2::buffer-group-description buffer-group)))
    (k-tabbar2::element-button
     (concat " " (k-tabbar2::buffer-group-name buffer-group) " ")
     description
     keymap
     (k-tabbar2::face k-tabbar2::button-size
		      text-color k-tabbar2::buffer-group-background-color nil)
     (k-tabbar2::face nil mouse-text-color 
		      k-tabbar2::buffer-group-background-color t)
     'buffer-group buffer-group
     )))



;; CACHE FUNCTIONS

(defun k-tabbar2::cache-get (symbol window)
  "Get the cached value for SYMBOL from the appropriate cache for WINDOW."
  (let ((cache-type (car (symbol-value symbol))))
    (cond
     ((eq cache-type 'window)
      (window-parameter window symbol))
     ((eq cache-type 'frame)
      (frame-parameter (window-frame window) symbol))
     (t
      (error "Unexpected cache-type (%s) for symbol %s"
	     cache-type symbol)))))

(defun k-tabbar2::cache-set (symbol value window)
  "Get the cached value for SYMBOL to VALUE within the appropriate cache
for WINDOW."
  (let ((cache-type (car (symbol-value symbol))))
    (cond
     ((eq cache-type 'window)
      (set-window-parameter window symbol value))
     ((eq cache-type 'frame)
      (set-frame-parameter (window-frame window) symbol value))
     (t
      (error "Unexpected cache-type (%s) for symbol %s"
	     cache-type symbol)))
    value))
       
(defun k-tabbar2::cache-reset (symbol window)
  "Clear the cached value for SYMBOL within the appropriate cache for
WINDOW."
  (k-tabbar2::cache-set symbol nil window))
  
(defun k-tabbar2::eval-symbol-fn (symbol window)
  "Call the function SYMBOL, with parameter WINDOW, which will return a
  value which is expected to subsequently be cached for that window." 
  (funcall (cadr (symbol-value symbol)) window))

(defun k-tabbar2::get (symbol window)
  "Get the value for symbol which may be cached or not, in the context
of WINDOW."
  (or (k-tabbar2::cache-get symbol window)
      (k-tabbar2::cache-set
       symbol (k-tabbar2::eval-symbol-fn symbol window) window)))


;; HIGH-LEVEL TABBAR ASSEMBLY FUNCTIONS AND CACHE DEFINITIONS

(defvar k-tabbar2::cached-cur-buffer
  '(window
   window-buffer)
  "Cache of window's last known selected buffer.  This is used to 
   figure out if the displayed buffer has changed, and so whether the
  tabbar should be updated.")

(defvar k-tabbar2::cached-buffer-group
  '(window
    k-tabbar2::default-buffer-group-for-cache)
  "Cache definition for a window's selected buffer-group.  This defines
  that the function `k-tabbar2::default-buffer-group-for-cache' is used
  by `k-tabbar2::get' to provide the value for this variable when it is
  nil.")

(defun k-tabbar2::default-buffer-group-for-cache (window)
  "Return the most appropriate buffer-group for WINDOW."
  (k-tabbar2::best-window-buffer-group window (window-buffer window)))

(defvar k-tabbar2::cached-buffers
  '(window
    k-tabbar2::tabbar-buffers-for-cache)
  "Cache definition for the buffers for a window's current buffer-group.")

(defun k-tabbar2::tabbar-buffers-for-cache (window)
  "Return a suitably ordered list of buffers for WINDOW's selected
buffer-group."
  (k-tabbar2::calculate-frame-buffer-group-buffer-list
   (window-frame window)
   (k-tabbar2::get 'k-tabbar2::cached-buffer-group window)))

(defvar k-tabbar2::cached-tabbar-type
  '(window
    k-tabbar2::default-tabbar-type-for-cache)
  "Window variable describing whether the tabbar is to display a
  buffer-list (if 'buffers) or buffer-groups (if 'buffer-groups) list.")

(defun k-tabbar2::default-tabbar-type-for-cache (window)
  "Return t.  This is a cache function used to initialise
  `k-tabbar2::cached-tabbar-type'."
  'buffers)

(defun k-tabbar2::remove-buffer-groups-for-other-frames (bgroups this-frame)
  "Return the list BGROUPS with frame buffer-groups for frames other
  than THIS-FRAME removed."
  (delq
   nil
   (mapcar
    (lambda (bgroup)
      (if (eq (car bgroup) 'frame)
	  (if (eq (cadr bgroup) this-frame)
	      bgroup)
	bgroup))
    bgroups)))

(defun k-tabbar2::buffer-groups-for-window (window)
  "Return a suitably ordered list of buffer-groups for WINDOW."
  (let* ((cur-bg (k-tabbar2::get 'k-tabbar2::cached-buffer-group window))
	 ;; The order of the following is important as
	 ;; k-tabbar2:all-buffer-groups may be updated by side-effect
	 ;; from evaluating `k-tabbar2::window-buffer-groups'
	 (win-bgs (k-tabbar2::window-buffer-groups window))
	 (shown-bgs (window-parameter window 'k-tabbar::buffer-groups))
	 sorted1
	 sorted2
	 result
	 )
    ;; Sort the list of all buffer-groups after removing the current
    ;; buffer-group and all frame buffer-groups for other frames.
    (setq sorted1 (k-tabbar2::sort-by-list
		   (k-tabbar2::remove-buffer-groups-for-other-frames
		    (delete cur-bg k-tabbar2::all-buffer-groups)
		    (window-frame window))
		   win-bgs))
    (setq sorted2 (k-tabbar2::sort-by-list
		   (cdr sorted1) shown-bgs))

    ;; (car sorted) now gives all bgs that have been explicitly selected
    ;; in this window, followed by those that have been implicitly
    ;; selected. (cdr sorted) gives the remainder.
    (setq result (append (car sorted1) (car sorted2))) 

    ;; Split (cdr sorted) into those buffer-groups that should be
    ;; highlighted for this window and those that should not.  The
    ;; highlighted ones will be appended to our result.
    (setq
     unhighlighted
     (delq nil
	   (mapcar
	    (lambda (buffer-group)
	      (if (k-tabbar2::buffer-group-known-to-frame
		   buffer-group (window-frame window))
		  (progn
		    (setq result (append result (list buffer-group)))
		    nil) ;; do not return this buffer-group
		buffer-group))
	    (car (cadr sorted2)))))
    (append result unhighlighted)))


(defvar k-tabbar2::cached-full-buffer-line
  '(window
    k-tabbar2::full-buffer-line-for-cache)
  "Cache definition for the buffers for a window's current buffer-group.")

(defun k-tabbar2::full-buffer-line-for-cache (window)
  "Calculate WINDOW's displayable full buffer-line.  This takes
  no account of buffer-name abbreviation or scrolling.  Each
  entry is its own list element (cons-cell).  The first entry is
  for the buffer-group.  The second is a separator entry.
  Subsequent entries alternate between buffer entries and
  separators with no final separator."
  (let ((buffer-group (k-tabbar2::get 'k-tabbar2::cached-buffer-group window))
	(buffer-list (k-tabbar2::get 'k-tabbar2::cached-buffers window))
	(separator (k-tabbar2::separator
		    k-tabbar2::buffer-background-color)))
    (apply
     'append
     (list
      (k-tabbar2::buffer-group-button
       k-tabbar2::buffer-group-keymap buffer-group window))
     (mapcar
      (lambda (buffer)
	(list separator
	      (k-tabbar2::buffer-button buffer window)))
      buffer-list))))

(defvar k-tabbar2::cached-buffer-line
  '(window
    k-tabbar2::buffer-line-for-cache)
  "Cache definition for the displayable set of buffer buttons for a
  window's current buffer-group, taking scrolling into account.") 

(defun k-tabbar2::buffer-line-for-cache (window)
  "Calculate WINDOW's displayable buffer-line, dealing with wrapping and
  scrolling as necessary."
  (let ((scroll-pos (k-tabbar2::get
		     'k-tabbar2::cached-scroll-position window)))
    (k-tabbar2::cond-add-right-scroll
     (k-tabbar2::cond-add-left-scroll
      (k-tabbar2::get 'k-tabbar2::cached-full-buffer-line window)
      k-tabbar2::buffer-background-color scroll-pos)
     k-tabbar2::buffer-background-color)))

(defvar k-tabbar2::cached-full-buffer-group-line
  '(window
    k-tabbar2::full-buffer-group-line-for-cache)
  "Cache definition for the window's current full buffer-group tabbar line.")

(defun k-tabbar2::full-buffer-group-line-for-cache (window)
  "Calculate WINDOW's displayable full buffer-group-line.
  This takes no account of scrolling.  Each entry is its own list
  element (cons-cell).  The first entry is for the current
  buffer-group.  The second is a separator entry.  Subsequent
  entries alternate between buffer group entries and separators
  with no final separator."
  (let ((cur-bgroup (k-tabbar2::get 'k-tabbar2::cached-buffer-group window))
	(bgroups (k-tabbar2::buffer-groups-for-window window))
	(separator (k-tabbar2::separator
		    k-tabbar2::buffer-group-background-color)))
    (apply
     'append
     (list
      (k-tabbar2::buffer-group-button
       k-tabbar2::buffer-group-keymap cur-bgroup window))
     (mapcar
      (lambda (buffer-group)
	(list separator
	      (k-tabbar2::buffer-group-button
	       k-tabbar2::buffer-group-entry-keymap
	       buffer-group window)))
      bgroups))))
	

(defvar k-tabbar2::cached-buffer-group-line
  '(window
    k-tabbar2::buffer-group-line-for-cache)
  "Cache definition for the displayable set of buffer-group buttons for
  a window's current buffer-group, taking scrolling into account.") 

(defun k-tabbar2::buffer-group-line-for-cache (window)
  "Calculate WINDOW's displayable buffer-group-line, dealing with
  wrapping and scrolling as necessary."
  (let ((scroll-pos (k-tabbar2::get
		      'k-tabbar2::cached-scroll-position window)))
    (k-tabbar2::cond-add-right-scroll
     (k-tabbar2::cond-add-left-scroll
      (k-tabbar2::get 'k-tabbar2::cached-full-buffer-group-line window)
      k-tabbar2::buffer-group-background-color scroll-pos)
     k-tabbar2::buffer-group-background-color)))

(defvar k-tabbar2::cached-scroll-position
  '(window
    k-tabbar2::zero-for-cache)
  "Cache definition for the scroll-position for a window's current
  buffer-group.")

(defun k-tabbar2::zero-for-cache (window)
  "Return zero.  This is a cache function used to initialise
  `k-tabbar2::cached-scroll-position'."
  0)

(defconst k-tabbar2::in-wrapped-buffer nil
  "This is redefined in dynamic scope when we are in a temporary
  buffer (used for determining line-wrapping).")

(defmacro k-tabbar2::with-wrap-buffer (&rest forms)
  "Run FORMS within a temp-buffer which can be used to determine
  whether text will fit a single line or be wrapped."
  `(with-temp-buffer
     ,(append
       '(let ((truncate-partial-width-windows)
	      (inhibit-modification-hooks t)
	      (truncate-lines nil)
	      (buffer-undo-list t)
	      (deactivate-mark) ;; Prevent deactivation of the mark!
	      (k-tabbar2::in-wrapped-buffer t)))
       forms
       )))

(defun k-tabbar2::init-wrap-buffer ()
  "Ensure that we are in a wrapped buffer and clear any existing text."
  (unless k-tabbar2::in-wrapped-buffer
    ;; This function must only be used from within
    ;; `k-tabbar2::with-wrap-buffer' 
    (error "Not in wrapped buffer!"))
  (delete-region (point-min) (point-max)))
  
(defun k-tabbar2::line-wrap-element (elements)
  "Returns the index of the element within the list, ELEMENTS (a list
  of insertable propertised text strings comprising a potential tabbar
  line), where text wrapping must occur.  IE, the last element from
  elements that can be displayed before being wrapped."
  (k-tabbar2::init-wrap-buffer)
  (let ((n 0))
    (catch 'wrap
      (mapc
       (lambda (elem)
	 (insert elem)
	 (unless (eq 'separator
		     (get-text-property 0 'element-type elem))
	   (goto-char 1)
	   (and (> (vertical-motion 1) 0)
		(throw 'wrap n))
	   (setq n (+ n 2))))
       elements)
      nil)))

(defun k-tabbar2::cond-add-left-scroll(source color scroll-pos)
  "Create a copy of the full buffer line given by SOURCE, removing
  leftmost elements according to the current window's scroll position,
  and adding a left scroll button with background color COLOR."
  (let ((elements (copy-sequence source)))
    (when (> scroll-pos 0)
      ;; Scroll-away the leftmost entries and add a left-scroll-button
      (let ((scroll-button (k-tabbar2::scroll-button '< color))
	    (sep (k-tabbar2::separator color)))
	(setcdr (nthcdr 1 elements)
		(cons scroll-button
		      (cons sep (nthcdr
				 (+ 2 (* 2 scroll-pos))
				 elements))))))
    elements))

(defun k-tabbar2::cond-add-right-scroll(buffer-line color)
  "Add a right scroll button, if needed, to BUFFER-LINE.
  TODO: Add a limit so that if no elements will appear in the
  buffer-line (maybe wrap < 4 or some such?) the buffer line is removed
  or replaced with a line of dots or something."
  (k-tabbar2::with-wrap-buffer
   (if (k-tabbar2::line-wrap-element buffer-line)
       (let* (;; Temporarily add the right-scroll button to buffer-list
	      ;; and figure out where wrapping really needs to occur.
	      (scroll-button (k-tabbar2::scroll-button '> color))
	      (tmp-buffer-line (cons scroll-button buffer-line))
	      (wrap (k-tabbar2::line-wrap-element tmp-buffer-line))
	      space)
	 ;; Truncate the buffer-line at wrap and spacing and the right
	 ;; scroll-button.
	 ;; SEE: https://www.emacswiki.org/emacs/HeaderLine for align-to
	 (setq
	  space
	  (propertize
	   " "
	   'display
	   `((space :align-to (- (+ right right-margin right-fringe) 4)))))
	 (setcdr (nthcdr (- wrap 3) buffer-line) (list space scroll-button))
	 )))
  buffer-line)

(defun k-tabbar2::element-type (str)
  "Return the element-type of propertized string STR."
  (get-text-property 0 'element-type str))

(defun k-tabbar2::tabbar-has-scroll-right-p (buffer-line)
  "t if BUFFER-LINE ends with a scroll-button."
  (let* ((len (length buffer-line))
	 (elem (nth (- len 1) buffer-line)))
    (eq (k-tabbar2::element-type elem) 'scroll-button)))
  

(defun k-tabbar2::last-selectable-tabbar-element (buffer-line)
  "Return the entry for the last selectable element in BUFFER-LINE."
  (let* ((len (length buffer-line))
	 (elem (nth (- len 1) buffer-line)))
    (if (eq (k-tabbar2::element-type elem) 'scroll-button)
	(nth (- len 3) buffer-line)
      elem)))

(defvar k-tabbar2::windows-to-redraw nil
  "A list of windows whose tabbars have been updated and should be
  redisplayed.  Actual redisplay will happen in the post-command hook.")

(defun k-tabbar2::queue-redisplay (window)
  "Queue a redisplay for WINDOW."
  (unless (member window k-tabbar2::windows-to-redraw)
    (setq k-tabbar2::windows-to-redraw
	  (cons window k-tabbar2::windows-to-redraw))))

(defun k-tabbar2::set-scroll-position (pos window)
  "Set the tabbar scroll position to POS for WINDOW.  This
  will cause tabbar-buffer-line to be recalculated."
  (k-tabbar2::cache-reset 'k-tabbar2::cached-buffer-line window)
  (k-tabbar2::cache-set 'k-tabbar2::cached-scroll-position pos window))

(defun k-tabbar2::scroll-right (window)
  "Action the tabbar right button in WINDOW."
  ;; UNTESTED BECAUSE UNCALLED
  (let ((last (k-tabbar2::last-selectable-tabbar-element
	       (k-tabbar2::get 'k-tabbar2::cached-buffer-line window)))
	(scroll-pos (k-tabbar2::cache-get
		     'k-tabbar2::cached-scroll-position window))
	tabbar)
    (k-tabbar2::set-scroll-position (1+ scroll-pos) window)
    (setq tabbar (k-tabbar2::get 'k-tabbar2::cached-buffer-line window))
    
    ;; If scrolling left has not affected the rightmost entry, do it again.
    (if (and (k-tabbar2::tabbar-has-scroll-right-p tabbar)
	     (eq last
		 (k-tabbar2::last-selectable-tabbar-element tabbar)))
	(k-tabbar2::scroll-right window))))
    
(defun k-tabbar2::scroll-left (window)
  "Action the tabbar left button in WINDOW."
  ;; UNTESTED BECAUSE UNCALLED
  (let ((scroll-pos (k-tabbar2::cache-get
		     'k-tabbar2::cached-scroll-position window)))
    (if (> scroll-pos 0)
	(k-tabbar2::set-scroll-position (1- scroll-pos) window))))
    
(defconst k-tabbar2::header-line-format
  '(:eval (k-tabbar2::tabbar (selected-window))))

(defun k-tabbar2::tabbar (window)
  "Generate a header-line entry for the current window and save the value in
window-parameter tabbar-line." 
  (if  k-tabbar2::windows-to-redraw
      ;; A redraw is currently queued.  Let's not do the recalculation
      ;; of the tabbar-line now.  Instead, we will get it from the
      ;; window parameter.  The required recalculation and redisplay
      ;; will happen later from our post-command-hook function.
      (or (window-parameter window 'tabbar-line))
    (set-window-parameter
     window 'tabbar-line
     (if (eq (k-tabbar2::get 'k-tabbar2::cached-tabbar-type window)
	     'buffers)
	   (k-tabbar2::get 'k-tabbar2::cached-buffer-line window)
       (k-tabbar2::get 'k-tabbar2::cached-buffer-group-line window)))))

    
;; Hooks and functions that make stuff happen.

(defun k-tabbar2::reset-window-state (window)
  "Clear all cache variables for WINDOW so that tabbar contents may be
  recalculated from scratch."
  (k-tabbar2::cache-reset 'k-tabbar2::cached-buffer-group window)
  (k-tabbar2::cache-reset 'k-tabbar2::cached-buffers window)
  (k-tabbar2::cache-reset 'k-tabbar2::cached-tabbar-type window)
  (k-tabbar2::cache-reset 'k-tabbar2::cached-buffer-line window)
  (k-tabbar2::cache-reset 'k-tabbar2::cached-full-buffer-line window)
  (k-tabbar2::cache-reset 'k-tabbar2::cached-buffer-group-line window)
  (k-tabbar2::cache-reset 'k-tabbar2::cached-full-buffer-group-line window))
  
(defun k-tabbar2::reset-state ()
  "Clear all cache variables so that tabbar contents are recalculated
  from scratch."
  (setq k-tabbar2::all-buffer-groups nil)
  ;; Ensure there is an emacs buffer-group
  (let ((b (or (get-buffer "*scratch*")
	       (get-buffer "*Messages*"))))
    (and b (k-tabbar2::buffer-group '(emacs))))
  (setq k-tabbar2::known-buffers nil)
  (walk-windows
   (lambda (window)
     (k-tabbar2::reset-window-state window)
     (k-tabbar2::cache-reset 'k-tabbar2::cached-cur-buffer window))
   nil t))

(defun k-tabbar2::resize-frame (frame)
  "Clear the appropriate caches when a frame's width changes so that the
  tabbar is appropriately resized."
  (walk-windows
   (lambda (window)
     (let ((width (window-pixel-width window)))
       (unless (eq width
		   (window-parameter window 'prev-width))
	 (set-window-parameter window 'prev-width width)
	 (message "Clear cache for %s" window)
	 (k-tabbar2::cache-reset 'k-tabbar2::cached-buffer-line window)
	 (k-tabbar2::cache-reset 'k-tabbar2::cached-buffer-group-line window)
	 (k-tabbar2::queue-redisplay window))))))

(defun k-tabbar2::remove-from-buffer-list (buffer window)
  "Remove BUFFER from the cached buffer list for WINDOW."
  (let ((buflist (k-tabbar2::get 'k-tabbar2::cached-buffers window)))
    (when (member buffer buflist)
      (k-tabbar2::cache-reset 'k-tabbar2::cached-buffer-line window)
      (k-tabbar2::cache-reset 'k-tabbar2::cached-full-buffer-line window)
      (k-tabbar2::cache-set
       'k-tabbar2::cached-buffers
       (delq buffer buflist)
       window)
      (k-tabbar2::queue-redisplay window))))

(defun k-tabbar2::forget-buffer-group (buffer-group)
  "Forget BUFFER-GROUP.  This means removing it from any displayed
buffer-group lists, and also from our list of all buffer-groups."
  (setq k-tabbar2::all-buffer-groups
	(delq buffer-group k-tabbar2::all-buffer-groups))
  (walk-windows
   (lambda (window)
     (when (eq (k-tabbar2::cache-get
		'k-tabbar2::cached-tabbar-type window)
	       'buffer-groups)
       ;; We need to redisplay this tabbar.
       (k-tabbar2::cache-reset
	'k-tabbar2::cached-full-buffer-group-line window)
       (k-tabbar2::cache-reset
	'k-tabbar2::cached-buffer-group-line window)
       (k-tabbar2::queue-redisplay window)))
   nil t))

(defun k-tabbar2::check-killed-buffer-groups (buffer)
  "Check whether we need to forget any buffer-groups as a result of
  BUFFER being killed.  We attempt to only do this for buffers that do
  not appear to be temp buffers."
  (unless (string-match-p "^ \\*" (buffer-name buffer))
    (mapc
     (lambda (buffer-group)
       (let ((buffers (k-tabbar2::buffer-group-buffers buffer-group)))
	 (when (and (= 1 (length buffers))
		    (eq (car buffers) buffer))
	   ;; Our buffer-group contains only buffer, so we can delete
	   ;; it.
	   (k-tabbar2::forget-buffer-group buffer-group))))
     (k-tabbar2::buffer-buffer-groups buffer))))

(defun k-tabbar2::forget-killed-buffer ()
  "Remove current-buffer from our caches and update whichever windows
  need updating."
  (k-tabbar2::check-killed-buffer-groups (current-buffer))
  (walk-windows
   (lambda (window)
     (when (eq (k-tabbar2::get'k-tabbar2::cached-tabbar-type window)
	       'buffers) 
       (k-tabbar2::remove-from-buffer-list (current-buffer) window)))
   nil t))
  
(defun k-tabbar2::add-to-buffer-list (buffer window prepend)
  "Add BUFFER to the cached buffer-list for WINDOW, adding to the start
  of the list if PREPEND, else to the end of the list."
  ;; TODO: Deal with scrolled buffer line
  (k-tabbar2::cache-reset 'k-tabbar2::cached-buffer-line window)
  (k-tabbar2::cache-reset 'k-tabbar2::cached-full-buffer-line window)
  (let ((buflist (k-tabbar2::get 'k-tabbar2::cached-buffers window)))
    (k-tabbar2::cache-set
     'k-tabbar2::cached-buffers
     (if prepend
	 (cons buffer buflist)
       (append buflist (list buffer)))
     window))
  (k-tabbar2::queue-redisplay window))

(defun k-tabbar2::handle-new-buffer (buffer)
  "Modify tabbars as appropriate for the new buffer BUFFER.  This is
  called when a new `current-buffer' is encountered from the
  buffer-update hook.  Note that the new buffer may cause new
  buffer-groups to be encountered." 
  (let* ((prev-bg-count (length k-tabbar2::all-buffer-groups))
	 (buffer-bgs (k-tabbar2::buffer-buffer-groups buffer))
	 (have-new-bg (not (= prev-bg-count
			      (length k-tabbar2::all-buffer-groups)))))
    (walk-windows
     (lambda (window)
       (let ((buffer-bgs (append buffer-bgs
				 (list (k-tabbar2::frame-buffer-group
					(window-frame window))))))
	 (if (eq (k-tabbar2::get'k-tabbar2::cached-tabbar-type window)
		 'buffers) 
	     (when (member
		    (k-tabbar2::cache-get
		     'k-tabbar2::cached-buffer-group window)
		    buffer-bgs)
	       ;; The new buffer *can* be shown in our buffer
	       ;; group.
	       (k-tabbar2::add-to-buffer-list
		buffer window 
		(eq window (selected-window))))
	   ;; tabbar is showing a buffer-group list
	   (when have-new-bg
	     (k-tabbar2::cache-reset
	      'k-tabbar2::cached-full-buffer-group-line window)
	     (k-tabbar2::cache-reset
	      'k-tabbar2::cached-buffer-group-line window)
	     (k-tabbar2::queue-redisplay window)))))
     nil t)
    nil))

(defvar k-tabbar2::buffers-updated nil
  "Boolean indicating that the buffer-list has changed and so we should 
check whether a window's selected buffer has changed, necessitating a
recalculation of the tabbar.")

(defun k-tabbar2::update-buffer-list ()
  "Called from the buffer-list update-hook.  We are only interested in
  whether we have a new current buffer.  Buffer deletions are handled by
  a separate function."
  (unless (minibufferp (current-buffer))
    (setq k-tabbar2::buffers-updated t)
    (when (k-tabbar2::new-buffer-p (current-buffer))
      (k-tabbar2::handle-new-buffer (current-buffer)))))

(defun k-tabbar2::pc-check-window-buffers ()
  "Function to run post-command to handle windows that are showing new
  buffers."
  (walk-windows
   (lambda (window)
     (unless (eq (window-buffer window)
		 (k-tabbar2::cache-get
		  'k-tabbar2::cached-cur-buffer window))
       ;; The window is showing a new buffer.
       ;; Ensure the buffer-group is appropriate
       (let ((prev-bg (k-tabbar2::cache-get
		       'k-tabbar2::cached-buffer-group window))
	     (bg (k-tabbar2::best-window-buffer-group
		  window (window-buffer window))))
	 (k-tabbar2::cache-set 'k-tabbar2::cached-buffer-group bg window)
	 (k-tabbar2::record-buffer-group-for-window window bg)
	 (k-tabbar2::cache-set
	  'k-tabbar2::cached-cur-buffer
	  (window-buffer window)
	  window)
	 (if (eq bg prev-bg)
	     ;; New buffer, but no change to buffer-group.  We need to
	     ;; recalculate the buffer-line to change the selected
	     ;; buffer.
	     (progn
	       (k-tabbar2::cache-reset
		'k-tabbar2::cached-buffer-line window)
	       (k-tabbar2::cache-reset
		'k-tabbar2::cached-full-buffer-line window))
	   ;; Ensure a redisplay will recalulate the tabbar-line
	   (k-tabbar2::reset-window-state window))
	 (k-tabbar2::queue-redisplay window))))
   nil t))
  
(defun k-tabbar2::pc-handle-queued-updates ()
  "Function to run post-command to redisplay windows that have queued updates.
  buffers."
  (mapc
   (lambda (window)
     (force-window-update window))
   k-tabbar2::windows-to-redraw)
  (setq k-tabbar2::windows-to-redraw nil))

(defun k-tabbar2::post-command-fn ()
  "Cause any queued window redisplays to happen."
  (when k-tabbar2::buffers-updated
    (k-tabbar2::pc-check-window-buffers)
    (setq k-tabbar2::buffers-updated nil))
  (k-tabbar2::pc-handle-queued-updates))

(defvar k-tabbar2::old-header-line-format nil
  "The previous header line value before we activated k-tabbar.")

(defun k-tabbar2::activate ()
  "Activate k-tabbar."
  (k-tabbar2::reset-state)
  (k-tabbar2::init-known-buffers)
  (k-tabbar2::init-cur-buffers)

  (walk-windows
   (lambda (window)
     (k-tabbar2::tabbar window)
     (set-window-parameter
      window 'k-tabbar::buffer-groups
      (list (k-tabbar2::cache-get 'k-tabbar2::cached-buffer-group window)))
     (k-tabbar2::queue-redisplay window))
   nil t)
  
  ;;(add-hook 'window-size-change-functions #'k-tabbar2::resize-frame)
  (add-hook 'buffer-list-update-hook #'k-tabbar2::update-buffer-list)
  (add-hook 'post-command-hook #'k-tabbar2::post-command-fn)
  (add-hook 'kill-buffer-hook #'k-tabbar2::forget-killed-buffer)
  (unless k-tabbar2::old-header-line-format
    (setq k-tabbar2::old-header-line-format header-line-format))
  (setq-default header-line-format k-tabbar2::header-line-format))

(defun k-tabbar2::deactivate ()
  "Deactivate k-tabbar."
  (remove-hook 'kill-buffer-hook #'k-tabbar2::forget-killed-buffer)
  (remove-hook 'post-command-hook #'k-tabbar2::post-command-fn)
  (remove-hook 'buffer-list-update-hook #'k-tabbar2::update-buffer-list)
  (remove-hook 'window-size-change-functions #'k-tabbar2::resize-frame)
  (setq-default header-line-format k-tabbar2::old-header-line-format)
  (setq k-tabbar2::old-header-line-format nil)
  (k-tabbar2::reset-state)
  (force-mode-line-update t))


;; BUTTON HELPER FUNCTIONS

(defun k-tabbar2::move-buffer-to-new-frame (buffer from-window ptr-pos)
  "Move BUFFER into its own frame, around the position of the pointer,
  removing it from the buffer-list of the frame for FROM-WINDOW.
  Attempt to move the frame to the pointer position PTR-POS."
  (let ((frame(make-frame-command)))
    (set-frame-position frame (car ptr-pos) (cdr ptr-pos))
    ;; TODO: Now forget buffer in the originating frame.
    ))
  

;; TABBAR BUTTON FUNCTIONS

(defun k-tabbar2::window-from-event (event)
  "Return the window from which EVENT was triggered."
  (caar (cdr event)))

(defun k-tabbar2::button-from-event (event)
  "Return the content of the button from where EVENT occurred."
  (nth 4 (cadr event)))

(defun k-tabbar2::buffer-from-event (event)
  "Return the buffer associated with the button where EVENT occurred."
  (let ((button (k-tabbar2::button-from-event event)))
    (get-text-property (cdr button) 'buffer (car button))))

(defun k-tabbar2::buffer-group-from-event (event)
  "Return the buffer-group associated with the button where EVENT occurred."
  (let ((button (k-tabbar2::button-from-event event)))
    (get-text-property (cdr button) 'buffer-group (car button))))

(defun k-tabbar2::select-buffer (event)
  "Select a buffer from a tabbar, determining the buffer from EVENT."
  (interactive "@e")
  (switch-to-buffer (k-tabbar2::buffer-from-event event)))
    
(defun k-tabbar2::select-buffer-group (event)
  "Select a buffer-group from a tabbar, determining the buffer-group
  from EVENT." 
  (interactive "@e")
  (let ((window (k-tabbar2::window-from-event event))
	(bgroup (k-tabbar2::buffer-group-from-event event)))
    (message "Setting bg to %s for %s" bgroup window)
    (k-tabbar2::reset-window-state window)
    (k-tabbar2::cache-set 'k-tabbar2::cached-buffer-group bgroup window) 
    (k-tabbar2::queue-redisplay window)))

(defun k-tabbar::get-drag-end-ptr-pos (event)
  "Return the pointer position from EVENT at the end of a mouse drag."
  ;; TODO: move to k-core
  (let* ((start-window (caar (cdr event)))
	 (frame-pos (frame-position (window-frame start-window)))
	 (buffer (k-tabbar::buffer-from-event event))
	 (event-end (event-end event))
	 (end-pos (nth 2 event-end)))
    (cons (+ (car end-pos) (car frame-pos))
	  (+ (cdr end-pos) (cdr frame-pos)))))

(defun k-tabbar2::get-drag-end-window (event)
  "Return the window to which the pointer has been dragged, based on
  EVENT.  If the pointer is not in an emacs window, return nil."
  ;; TODO: move to k-core
  (let ((event-end (event-end event)))
    (or (windowp (car event-end))
	(k::window-for-position (k-tabbar2::get-drag-end-ptr-pos event)))))
  
(defun k-tabbar2::buffer-drag (event)
  "Handle buffer placement by mouse drag."
  (interactive "@e")
  (let ((start-window (caar (cdr event)))
	(end-window (k-tabbar2::get-drag-end-window event)))
    (if end-window
	;; Handle the end-event being a window on the same frame
	(unless (eq start-window end-window)
	  (k-tabbar2::move-buffer buffer start-window end-window))
      (k-tabbar2::move-buffer-to-new-frame
       (k-tabbar::buffer-from-event event)
       start-window
       (k-tabbar2::get-drag-end-ptr-pos event)))))
    
(defun k-tabbar2::buffer-drag-old (event)
  "Handle buffer placement by mouse drag."
  (interactive "@e")
  (let ((start-window (caar (cdr event)))
	(event-end (event-end event))
	end-window)
    (if (setq end-window (windowp (car event-end)))
	;; Handle the end-event being a window on the same frame
	(unless (eq start-window end-window)
	  (k-tabbar2::move-buffer buffer start-window end-window))
      (let* ((frame-pos (frame-position (window-frame start-window)))
	     (buffer (k-tabbar::buffer-from-event event))
	     (end-pos (nth 2 event-end))
	     (ptr-pos (cons (+ (car end-pos) (car frame-pos))
			    (+ (cdr end-pos) (cdr frame-pos))))
	     (end-window (k::window-for-position ptr-pos)))
	(if end-window
	    (k-tabbar2::move-buffer buffer start-window end-window)
	  (k-tabbar2::move-buffer-to-new-frame buffer start-window))))))
    

(defun k-tabbar2::act-on-buffer-entry (event)
  "Act on EVENT, which will be a key-click from a buffer button."
  (interactive "@e")
  (let ((type (car event))
	(window (caar (cdr event)))
	(buffer (k-tabbar::buffer-from-event event))
	
    (message "EVENT TYPE: %s, in window %s, for buffer %s" type window
	     buffer)
    (message "EVENT: %s" event)
    (when (eq type 'drag-mouse-1)
      (message "DRAG FROM WINDOW: %s" window)
      (let ((end (event-end event))
	    (frame-pos (frame-position (window-frame window)))
	    end-pos)
	(message "End: %s" end)
	(setq end-pos
	      (cons (+ (car (nth 2 end)) (car frame-pos))
		    (+ (cdr (nth 2 end)) (cdr frame-pos))))
	(message "end-pos: %s" end-pos)
	(unless (windowp (car end))
	  (message "DRAGGED TO WINDOW: %s"
		   (k::window-for-position end-pos))))))))
      

	
(defun k-tabbar2::toggle-tabbar-display-type (event)
  "Switch the mode of the tabbar line between display of buffer and
buffer-groups."
  (interactive "@e")
  (let ((window (k-tabbar2::window-from-event event)))
    (if (eq (k-tabbar2::get 'k-tabbar2::cached-tabbar-type window) 'buffers)
	(k-tabbar2::cache-set
	 'k-tabbar2::cached-tabbar-type 'buffer-groups window)
      (k-tabbar2::cache-set
       'k-tabbar2::cached-tabbar-type 'buffers window))
    (k-tabbar2::cache-reset 'k-tabbar2::cached-buffer-line window)
    (k-tabbar2::cache-reset 'k-tabbar2::cached-buffer-group-line window)
    (k-tabbar2::queue-redisplay window)))



;; Cached values.
;; These can be retrieved using (k-tabbar2::get)
;;
;; k-tabbar2::cached-buffer-group (per window)
;;   The currently selected buffer-group for the window
;;
;; k-tabbar2::cached-buffers (per window)
;;   A suitably ordered list of buffers for the current window's
;;   buffer-group (as a list suitable for display in the header line).
;;
    ;; k-tabbar2::cached-full-buffer-line (per window)
;;   The full tabbar line for buffers without consideration of name
;;   abbreviations or scrolling (as a list suitable for display in the
;;   header line). 
;;
;; k-tabbar2::cached-buffer-line (per window)
;;   The tabbar line for buffers, set to the scroll-position,
;;   truncated as necessary and with whatever scroll buttons are needed
;;   (as a list suitable for display in the header line). 
;;   
;; k-tabbar2::cached-scroll-position (per window)
;;   The scroll-position (how many elements have scrolled-off to the
;;   left) for the current window's tabbar-buffer-line or
;;   tabbar-buffer-groups-line, whichever is selected.
;;
;; k-tabbar2::tabbar-alt-scroll-position (per window)
;;   The scroll-position (how many elements have scrolled-off to the
;;   left) for the current window's tabbar-buffer-line or
;;   tabbar-buffer-groups-line, whichever is NOT selected.
;;

;; TODO:
;; 1) Scrolling
;;    - Add scroll button functionality
;;    - Add scroll button help text
;;    - Make frame resize recalculate the display-line
;;    - Ensure this works in both buffer and buffer-group displays
;; 2) Exotic functionality (drag and drop, etc?)
;; 3) Rationalise/complete keymaps
;; 4) replace for existing tabbar code
;; 5) Header comments
;; 6) any inline TODOs


(progn
  (k-tabbar2::deactivate)
  (k-tabbar2::activate)
  (message "x"))

(provide 'k-tabbar2)
